from __future__ import absolute_import

from datetime import datetime

import atomize

from feedmark.feeds import extract_feed_properties, extract_sections, construct_entry_url
from feedmark.formats.markdown import markdown_to_html5


def convert_section_to_entry(section, properties, markdown_links_base=None):

    guid = properties['url'] + "/" + section.title
    updated = section.publication_date

    summary = atomize.Summary(markdown_to_html5(section.body), content_type='html')

    links = []
    entry_url = construct_entry_url(section)
    if entry_url is not None:
        links.append(atomize.Link(entry_url, content_type='text/html', rel='alternate'))

    return atomize.Entry(
        title=section.title,
        guid=guid,
        updated=updated,
        summary=summary,
        links=links
    )


def feedmark_atomize(documents, out_filename, limit=None):
    properties = {}

    for document in documents:
        these_properties = extract_feed_properties(document)
        properties.update(these_properties)  # TODO: something more elegant than this

    entries = []
    for section in extract_sections(documents):
        entries.append(convert_section_to_entry(section, properties))

    if limit and len(entries) > limit:
        entries = entries[:limit]

    assert properties['author'], "Need author"

    feed = atomize.Feed(
        author=properties['author'],
        title=properties['title'],
        updated=datetime.utcnow(),
        guid=properties['url'],
        self_link=properties['url'],
        entries=entries
    )

    feed.write_file(out_filename)



# Python 2/3
try:
    unicode = unicode
except NameError:
    unicode = str

try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO
assert StringIO

try:
    from urllib import quote, quote_plus
except ImportError:
    from urllib.parse import quote, quote_plus
assert quote and quote_plus


def items(d):
    try:
        return d.iteritems()
    except AttributeError:
        return d.items()


def items_in_priority_order(di, priority):
    for key in priority:
        if key in di:
            yield key, di[key]
    for key, item in sorted(items(di)):
        if key not in priority:
            yield key, item



from argparse import ArgumentParser
import json
import sys

from feedmark.loader import (
    read_document_from, read_refdex_from, convert_refdex_to_single_filename_refdex,
)
from feedmark.utils import items


def sys_main():
    return main(sys.argv[1:])


def main(args):
    argparser = ArgumentParser()

    argparser.add_argument('input_files', nargs='+', metavar='FILENAME', type=str,
        help='Markdown files containing the embedded entries'
    )

    argparser.add_argument('--by-property', action='store_true',
        help='Output JSON containing a list of all properties found and the entries they were found on'
    )
    argparser.add_argument('--by-publication-date', action='store_true',
        help='Output JSON list of the embdedded entries, sorted by publication date'
    )
    argparser.add_argument('--dump-entries', action='store_true',
        help='Output indented summary of the entries on standard output'
    )
    argparser.add_argument('--output-json', action='store_true',
        help='Output JSON containing entries on standard output'
    )
    argparser.add_argument('--htmlized-json', action='store_true',
        help='When outputting JSON, convert Markdown fields (preamble, section bodies, etc) to HTML5'
    )
    argparser.add_argument('--ordered-json', action='store_true',
        help='When outputting JSON, generate properties as lists that preserve the order '
             'from the source Feedmark document, instead of as unordered objects'
    )

    argparser.add_argument('--output-links', action='store_true',
        help='Output JSON containing all web links extracted from the entries'
    )

    argparser.add_argument('--check-against-schema', metavar='FILENAME', type=str, default=None,
        help='Check if entries have the properties specified by this schema.  This schema will '
             'also provide hints (such as ordering of properties) when outputting Markdown or HTML.'
    )

    argparser.add_argument('--output-atom', metavar='FILENAME', type=str,
        help='Construct an Atom XML feed from the entries and write it out to this file'
    )
    argparser.add_argument('--output-markdown', action='store_true',
        help='Reconstruct a Markdown document from the entries and write it to stdout'
    )
    argparser.add_argument('--output-html', action='store_true',
        help='Construct an HTML5 article element from the entries and write it to stdout'
    )

    argparser.add_argument('--rewrite-markdown', action='store_true',
        help='Rewrite all input Markdown documents in-place. Note!! Destructive!!'
    )

    argparser.add_argument('--input-refdex', metavar='FILENAME', type=str,
        help='Load this JSON file as the reference-style links index before processing'
    )
    argparser.add_argument('--input-refdexes', metavar='FILENAME', type=str,
        help='Load these JSON files as the reference-style links index before processing'
    )
    argparser.add_argument('--input-refdex-filename-prefix', type=str, default=None,
        help='After loading refdexes, prepend this to filename of each refdex'
    )
    argparser.add_argument('--output-refdex', action='store_true',
        help='Construct reference-style links index from the entries and write it to stdout as JSON'
    )
    argparser.add_argument('--output-refdex-single-filename', action='store_true',
        help='When outputting a refdex, ensure that only entries with a single filename are '
             'output, by stripping all but the last filename from multiple filenames entries.'
    )

    argparser.add_argument('--limit', metavar='COUNT', type=int, default=None,
        help='Process no more than this many entries when making an Atom or HTML feed'
    )

    argparser.add_argument('--version', action='version', version="%(prog)s 0.14")

    options = argparser.parse_args(args)

    documents = []

    ### input

    for filename in options.input_files:
        document = read_document_from(filename)
        documents.append(document)

    ### input: load input refdexes

    input_refdexes = []
    if options.input_refdex:
        input_refdexes.append(options.input_refdex)
    if options.input_refdexes:
        for input_refdex in options.input_refdexes.split(','):
            input_refdexes.append(input_refdex.strip())

    refdex = read_refdex_from(input_refdexes, input_refdex_filename_prefix=options.input_refdex_filename_prefix)

    ### processing

    schema = None
    if options.check_against_schema is not None:
        from feedmark.checkers import Schema
        schema_document = read_document_from(options.check_against_schema)
        schema = Schema(schema_document)
        results = schema.check_documents(documents)
        if results:
            sys.stdout.write(json.dumps(results, indent=4, sort_keys=True))
            sys.exit(1)

    ### processing: collect refdex phase
    # NOTE: we only run this if we were asked to output a refdex -
    # this is to prevent scurrilous insertion of refdex entries when rewriting.

    if options.output_refdex:
        for document in documents:
            for section in document.sections:
                if section.title in refdex:
                    entry = refdex[section.title]
                    if entry['anchor'] != section.anchor:
                        raise ValueError("Inconsistent anchors: {} in refex, {} in document".format(entry['anchor'], section.anchor))
                    if 'filename' in entry:
                        entry['filenames'] = []
                        del entry['filename']
                    entry['filenames'].append(document.filename)
                else:
                    refdex[section.title] = {
                        'filenames': [document.filename],
                        'anchor': section.anchor
                    }

    ### processing: rewrite references phase

    if refdex:
        for document in documents:
            document.rewrite_reference_links(refdex)

    ### output

    if options.output_refdex:
        if options.output_refdex_single_filename:
            refdex = convert_refdex_to_single_filename_refdex(refdex)
        sys.stdout.write(json.dumps(refdex, indent=4, sort_keys=True))

    if options.dump_entries:
        for document in documents:
            for section in document.sections:
                print(section.title)
                for (name, url) in section.images:
                    print(u'    !{}: {}'.format(name, url))
                for key, value in items(section.properties):
                    if isinstance(value, list):
                        print(u'    {}@'.format(key))
                        for subitem in value:
                            print(u'        {}'.format(subitem))
                    else:
                        print(u'    {}: {}'.format(key, value))

    if options.output_json:
        json_options = {
            'htmlize': options.htmlized_json,
            'ordered': options.ordered_json,
        }
        output_json = {
            'documents': [d.to_json_data(**json_options) for d in documents]
        }
        sys.stdout.write(json.dumps(output_json, indent=4, sort_keys=True))

    if options.by_publication_date:
        from feedmark.feeds import construct_entry_url

        dated_items = []
        for document in documents:
            for section in document.sections:
                section_json = {
                    'title': section.title,
                    'images': section.images,
                    'properties': section.properties,
                    'body': section.body,
                    'url': construct_entry_url(section)
                }
                dated_items.append((section.publication_date, section_json))
        dated_items.sort(reverse=True)
        if options.limit:
            dated_items = dated_items[:options.limit]
        output_json = [item for (d, item) in dated_items]
        sys.stdout.write(json.dumps(output_json, indent=4, sort_keys=True))

    if options.by_property:
        by_property = {}
        for document in documents:
            for section in document.sections:
                for key, value in items(section.properties):
                    if isinstance(value, list):
                        key = u'{}@'.format(key)
                    by_property.setdefault(key, {}).setdefault(section.title, value)
        sys.stdout.write(json.dumps(by_property, indent=4))

    if options.output_links:
        from feedmark.checkers import extract_links_from_documents
        links = extract_links_from_documents(documents)
        sys.stdout.write(json.dumps(links, indent=4, sort_keys=True))

    if options.output_markdown:
        from feedmark.formats.markdown import feedmark_markdownize
        for document in documents:
            s = feedmark_markdownize(document, schema=schema)
            sys.stdout.write(s)

    if options.rewrite_markdown:
        from feedmark.formats.markdown import feedmark_markdownize
        for document in documents:
            s = feedmark_markdownize(document, schema=schema)
            with open(document.filename, 'w') as f:
                f.write(s)

    if options.output_html:
        from feedmark.formats.markdown import feedmark_htmlize
        for document in documents:
            s = feedmark_htmlize(document, schema=schema)
            sys.stdout.write(s)

    if options.output_atom:
        from feedmark.formats.atom import feedmark_atomize
        feedmark_atomize(documents, options.output_atom, limit=options.limit)


if __name__ == '__main__':
    main(sys.argv[1:])



import codecs
import json
import sys

from feedmark.parser import Parser
from feedmark.utils import items


def read_document_from(filename):
    with codecs.open(filename, 'r', encoding='utf-8') as f:
        markdown_text = f.read()
    parser = Parser(markdown_text)
    document = parser.parse_document()
    document.filename = filename
    return document


def read_refdex_from(filenames, input_refdex_filename_prefix=None):
    refdex = {}
    for filename in filenames:
        try:
            with codecs.open(filename, 'r', encoding='utf-8') as f:
                local_refdex = json.loads(f.read())
                if input_refdex_filename_prefix:
                    for key, value in items(local_refdex):
                        if 'filename' in value:
                            value['filename'] = input_refdex_filename_prefix + value['filename']
                        if 'filenames' in value:
                            value['filenames'] = [input_refdex_filename_prefix + f for f in value['filenames']]
                refdex.update(local_refdex)
        except:
            sys.stderr.write("Could not read refdex JSON from '{}'\n".format(filename))
            raise

    # Python 2/3
    try:
        unicode_string = unicode
    except NameError:
        unicode_string = str

    for key, value in items(refdex):
        try:
            assert isinstance(key, unicode_string)
            if 'url' in value:
                assert len(value) == 1
                assert isinstance(value['url'], unicode_string)
                value['url'].encode('utf-8')
            elif 'filename' in value and 'anchor' in value:
                assert len(value) == 2
                assert isinstance(value['filename'], unicode_string)
                value['filename'].encode('utf-8')
                assert isinstance(value['anchor'], unicode_string)
                value['anchor'].encode('utf-8')
            elif 'filenames' in value and 'anchor' in value:
                assert len(value) == 2
                for filename in value['filenames']:
                    assert isinstance(value, unicode_string)
                    filename.encode('utf-8')
                assert isinstance(value['anchor'], unicode_string)
                value['anchor'].encode('utf-8')
            else:
                raise NotImplementedError("badly formed refdex")
        except:
            sys.stderr.write("Component of refdex not suitable: '{}: {}'\n".format(repr(key), repr(value)))
            raise

    return refdex


def convert_refdex_to_single_filename_refdex(input_refdex):
    """Note that this makes a partially shallow copy."""
    refdex = {}
    for key, value in input_refdex.items():
        if 'filenames' in value:
            refdex[key] = {
                'filename': value['filenames'][-1],
                'anchor': value['anchor']
            }
        else:
            refdex[key] = value
    return refdex



import unittest

import json
import os
import sys
from subprocess import check_call
from tempfile import mkdtemp

from feedmark.checkers import Schema
from feedmark.main import main
from feedmark.loader import read_document_from
from feedmark.utils import StringIO


class TestFeedmarkFileCreation(unittest.TestCase):

    def setUp(self):
        super(TestFeedmarkFileCreation, self).setUp()
        self.saved_stdout = sys.stdout
        sys.stdout = StringIO()
        self.maxDiff = None
        self.dirname = mkdtemp()
        self.prevdir = os.getcwd()
        os.chdir(self.dirname)

    def tearDown(self):
        os.chdir(self.prevdir)
        check_call("rm -rf {}".format(self.dirname), shell=True)
        sys.stdout = self.saved_stdout
        super(TestFeedmarkFileCreation, self).tearDown()

    def assert_file_contains(self, filename, text):
        with open(filename, 'r') as f:
            contents = f.read()
        self.assertIn(text, contents)

    def test_atom_feed(self):
        main(["{}/eg/Recent Llama Sightings.md".format(self.prevdir), '--output-atom=feed.xml'])
        self.assert_file_contains('feed.xml', '<id>http://example.com/llama.xml/2 Llamas Spotted Near Mall</id>')
        self.assert_file_contains('feed.xml',
            'https://github.com/catseye/Feedmark/blob/master/eg/Recent%20Llama%20Sightings.md#2-llamas-spotted-near-mall'
        )
        os.unlink('feed.xml')

    def test_rewrite_markdown_input_refdex(self):
        with open('foo.md', 'w') as f:
            f.write("""# Document

### Entry

Have you heard, [2 Llamas Spotted Near Mall]()?

[2 Llamas Spotted Near Mall]: TK
""")
        main(["foo.md", "--input-refdex={}/eg/refdex.json".format(self.prevdir), '--rewrite-markdown'])
        self.assert_file_contains('foo.md', '[2 Llamas Spotted Near Mall]: eg/Recent%20Llama%20Sightings.md#2-llamas-spotted-near-mall')
        os.unlink('foo.md')

    def test_rewrite_markdown_internal(self):
        with open('foo.md', 'w') as f:
            f.write("""# Document

### Bubble & Squeak

Have you heard, [Bubble & Squeak]()?

[Bubble & Squeak]: TK
""")
        main(["foo.md", '--output-refdex', '--rewrite-markdown'])
        self.assert_file_contains('foo.md', '[Bubble & Squeak]: foo.md#bubble--squeak')
        os.unlink('foo.md')


class TestFeedmarkCommandLine(unittest.TestCase):

    def setUp(self):
        super(TestFeedmarkCommandLine, self).setUp()
        self.saved_stdout = sys.stdout
        sys.stdout = StringIO()
        self.maxDiff = None

    def tearDown(self):
        sys.stdout = self.saved_stdout
        super(TestFeedmarkCommandLine, self).tearDown()

    def test_schema(self):
        main(["eg/Recent Llama Sightings.md", "eg/Ancient Llama Sightings.md", '--check-against-schema=eg/schema/Llama sighting.md'])
        output = sys.stdout.getvalue()
        self.assertEqual(output, '')

    def test_schema_failure(self):
        with self.assertRaises(SystemExit):
            main(["eg/Ill-formed Llama Sightings.md", "eg/Recent Llama Sightings.md", '--check-against-schema=eg/schema/Llama sighting.md'])
        data = json.loads(sys.stdout.getvalue())
        self.assertEqual(data, [
            {
                u'document': u'Ill-formed Llama Sightings',
                u'result': [[u'extra', u'excuse'], [u'missing', u'date']],
                u'section': u'Definite llama sighting with no date'
            }
        ])

    def test_output_html(self):
        main(["eg/Recent Llama Sightings.md", "--output-html"])
        output = sys.stdout.getvalue()
        self.assertIn('<h3 id="a-possible-llama-under-the-bridge">A Possible Llama Under the Bridge</h3>', output)

    def test_output_json(self):
        main(['eg/Ancient Llama Sightings.md', '--output-json'])
        data = json.loads(sys.stdout.getvalue())
        self.assertDictEqual(data, {
            u'documents': [
                {
                    u'filename': u'eg/Ancient Llama Sightings.md',
                    u'title': u'Ancient Llama Sightings',
                    u'preamble': u'',
                    u'properties': data['documents'][0]['properties'],
                    u'sections': data['documents'][0]['sections'],
                }
            ]
        })
        self.assertDictEqual(data['documents'][0]['properties'], {
            u'author': u'Alfred J. Prufrock',
            u'link-target-url': u'https://github.com/catseye/Feedmark/blob/master/eg/Ancient%20Llama%20Sightings.md',
            u'url': u'http://example.com/old_llama.xml'
        })
        self.assertEqual(data['documents'][0]['sections'], [
            {
                u'body': data['documents'][0]['sections'][0]['body'],
                u'images': [
                    {
                        u'description': u'photo of possible llama',
                        u'source': u'https://static.catseye.tc/images/screenshots/Kolakoski_Kurve.jpg',
                    }
                ],
                u'properties': {u'date': u'Jan 1 1984 12:00:00'},
                u'title': u'Maybe sighting the llama',
                u'anchor': u'maybe-sighting-the-llama',
            }
        ])
        self.assertIn(u'It was a possible llama sighting.\n\n', data['documents'][0]['sections'][0]['body'])

    def test_output_json_with_multiple_images_and_linked_images(self):
        main(['eg/Recent Llama Sightings.md', '--output-json'])
        data = json.loads(sys.stdout.getvalue())
        self.assertEqual(data['documents'][0]['sections'][1]['images'], [
            {
                u'description': u'photo of possible llama',
                u'source': u'https://static.catseye.tc/images/screenshots/Heronsis_hermnonicii.jpg',
                u'link': u'https://catseye.tc/article/Gewgaws.md',
            },
            {
                u'description': u'another possible photo',
                u'source': u'https://static.catseye.tc/images/screenshots/A_Non-Random_Walk.jpg',
            },
        ])

    def test_output_htmlized_json(self):
        main(['eg/Referenced Llama Sightings.md', '--output-json', '--htmlized-json'])
        data = json.loads(sys.stdout.getvalue())
        self.assertDictEqual(data, {
            u'documents': [
                {
                    u'filename': u'eg/Referenced Llama Sightings.md',
                    u'title': u'Referenced Llama Sightings',
                    u'preamble': u'<p>Some <strong>llamas</strong> have been <a href="spotted.html">spotted</a> recently.</p>',
                    u'properties': data['documents'][0]['properties'],
                    u'sections': data['documents'][0]['sections'],
                }
            ]
        })
        self.assertEqual(
            data['documents'][0]['sections'][0]['body'],
            u"<p>I have strong opinions about this.  It's a <em>shame</em> more llamas aren't\nbeing spotted.  "
             "Sometimes they are <strong>striped</strong>, it's true, but<br />\nwhen<br />\nthey are, "
             "<a href=\"https://daringfireball.net/projects/markdown/\">Markdown</a>\ncan be used.</p>\n"
             "<p>To <a href=\"https://en.wikipedia.org/wiki/Site\">site</a> them.</p>\n<p>Sight them, sigh.</p>"
        )
        # note that property values are bare HTML fragments: there is no surrounding <p></p> or other element
        self.assertEqual(
            data['documents'][0]['properties']['hopper'],
            '<a href="https://en.wikipedia.org/wiki/Stephen_Hopper">Stephen</a>'
        )
        self.assertEqual(
            data['documents'][0]['properties']['spotted'],
            [u'<a href="mall.html">the mall</a>', u'<a href="beach.html">the beach</a>']
        )
        self.assertEqual(
            data['documents'][0]['sections'][0]['properties']['hopper'],
            '<a href="https://en.wikipedia.org/wiki/Grace_Hopper">Grace</a>'
        )
        self.assertEqual(
            data['documents'][0]['sections'][0]['properties']['spotted'],
            [u'<a href="mall.html">the mall</a>', u'<a href="lumberyard.html">the lumberyard</a>']
        )

    def test_output_unordered_json(self):
        main(['eg/Referenced Llama Sightings.md', '--output-json'])
        data = json.loads(sys.stdout.getvalue())
        self.assertDictEqual(data['documents'][0]['properties'], {
            u'author': u'Alfred J. Prufrock',
            u'link-target-url': u'https://github.com/catseye/Feedmark/blob/master/eg/Referenced%20Llama%20Sightings.md',
            u'url': u'http://example.com/refllama.xml',
            u'hopper': u'[Stephen](https://en.wikipedia.org/wiki/Stephen_Hopper)',
            u'spotted': [u'[the mall][]', u'[the beach](beach.html)'],
        })
        self.assertDictEqual(data['documents'][0]['sections'][0]['properties'], {
            u'date': u'Nov 1 2016 09:00:00',
            u'hopper': u'[Grace](https://en.wikipedia.org/wiki/Grace_Hopper)',
            u'spotted': [u'[the mall][]', u'[the lumberyard](lumberyard.html)'],
        })

    def test_output_ordered_json(self):
        main(['eg/Referenced Llama Sightings.md', '--output-json', '--ordered-json'])
        data = json.loads(sys.stdout.getvalue())
        self.assertEqual(data['documents'][0]['properties'], [
            [u'author', u'Alfred J. Prufrock'],
            [u'url', u'http://example.com/refllama.xml'],
            [u'link-target-url', u'https://github.com/catseye/Feedmark/blob/master/eg/Referenced%20Llama%20Sightings.md'],
            [u'hopper', u'[Stephen](https://en.wikipedia.org/wiki/Stephen_Hopper)'],
            [u'spotted', [u'[the mall][]', u'[the beach](beach.html)']],
        ])
        self.assertEqual(data['documents'][0]['sections'][0]['properties'], [
            [u'date', u'Nov 1 2016 09:00:00'],
            [u'hopper', u'[Grace](https://en.wikipedia.org/wiki/Grace_Hopper)'],
            [u'spotted', [u'[the mall][]', u'[the lumberyard](lumberyard.html)']]
        ])

    def test_output_refdex(self):
        main(['eg/Recent Llama Sightings.md', 'eg/Ancient Llama Sightings.md', '--output-refdex'])
        data = json.loads(sys.stdout.getvalue())
        self.assertDictEqual(data, {
            "2 Llamas Spotted Near Mall": {
                "anchor": "2-llamas-spotted-near-mall",
                "filenames": ["eg/Recent Llama Sightings.md"],
            },
            "A Possible Llama Under the Bridge": {
                "anchor": "a-possible-llama-under-the-bridge",
                "filenames": ["eg/Recent Llama Sightings.md"],
            },
            "Llamas: It's Time to Spot Them": {
                "anchor": "llamas-its-time-to-spot-them",
                "filenames": ["eg/Recent Llama Sightings.md"],
            },
            "Maybe sighting the llama": {
                "anchor": "maybe-sighting-the-llama",
                "filenames": ["eg/Ancient Llama Sightings.md"],
            }
        })

    def test_output_refdex_with_overlap(self):
        # Both of these files contain an entry called "Llamas: It's Time to Spot Them".
        # The refdex is created with entries pointing to all files where the entry occurs.
        main(['eg/Recent Llama Sightings.md', 'eg/Referenced Llama Sightings.md', '--output-refdex'])
        data = json.loads(sys.stdout.getvalue())
        self.assertDictEqual(data, {
            "2 Llamas Spotted Near Mall": {
                "anchor": "2-llamas-spotted-near-mall",
                "filenames": [
                    "eg/Recent Llama Sightings.md",
                ]
            },
            "A Possible Llama Under the Bridge": {
                "anchor": "a-possible-llama-under-the-bridge",
                "filenames": [
                    "eg/Recent Llama Sightings.md",
                ],
            },
            "Llamas: It's Time to Spot Them": {
                "anchor": "llamas-its-time-to-spot-them",
                "filenames": [
                    "eg/Recent Llama Sightings.md",
                    "eg/Referenced Llama Sightings.md"
                ]
            },
        })

    def test_output_refdex_with_overlap_forcing_single_filename(self):
        # Both of these files contain an entry called "Llamas: It's Time to Spot Them"
        # The refdex is created pointing only to the file that was mentioned last.
        main(['eg/Recent Llama Sightings.md', 'eg/Referenced Llama Sightings.md', '--output-refdex', '--output-refdex-single-filename'])
        data = json.loads(sys.stdout.getvalue())
        self.assertDictEqual(data, {
            "2 Llamas Spotted Near Mall": {
                "anchor": "2-llamas-spotted-near-mall",
                "filename": "eg/Recent Llama Sightings.md",
            },
            "A Possible Llama Under the Bridge": {
                "anchor": "a-possible-llama-under-the-bridge",
                "filename": "eg/Recent Llama Sightings.md",
            },
            "Llamas: It's Time to Spot Them": {
                "anchor": "llamas-its-time-to-spot-them",
                "filename": "eg/Referenced Llama Sightings.md",
            },
        })

    def test_input_refdex_output_markdown(self):
        main(['eg/Ill-formed Llama Sightings.md', '--input-refdex', 'eg/refdex.json', '--output-markdown'])
        output = sys.stdout.getvalue()
        self.assertIn('[2 Llamas Spotted Near Mall]: eg/Recent%20Llama%20Sightings.md#2-llamas-spotted-near-mall', output)

    def test_output_links(self):
        main(['eg/Ill-formed Llama Sightings.md', '--output-links'])
        data = json.loads(sys.stdout.getvalue())
        self.assertEqual(data, [
            {
                u'document': u'Ill-formed Llama Sightings',
                u'name': u'2 Llamas Spotted Near Mall',
                u'section': u'Definite llama sighting with no date',
                u'url': u'TK'
            },
            {
                u'document': u'Ill-formed Llama Sightings',
                u'section': u'Definite llama sighting with no date',
                u'url': u'https://tcrf.net/The_Cutting_Room_Floor'
            }
        ])


class TestFeedmarkInternals(unittest.TestCase):

    def test_load_documents(self):
        doc1 = read_document_from('eg/Ancient Llama Sightings.md')
        self.assertEqual(doc1.title, "Ancient Llama Sightings")
        doc2 = read_document_from('eg/Recent Llama Sightings.md')
        self.assertEqual(doc2.title, "Recent Llama Sightings")
        self.assertEqual(len(doc2.sections), 3)

    def test_schema(self):
        schema_doc = read_document_from('eg/schema/Llama sighting.md')
        schema = Schema(schema_doc)

        doc1 = read_document_from('eg/Ancient Llama Sightings.md')
        doc2 = read_document_from('eg/Recent Llama Sightings.md')
        results = schema.check_documents([doc1, doc2])
        self.assertEqual(results, [])


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
# encoding: UTF-8

from datetime import datetime
from collections import OrderedDict
import re

from feedmark.formats.markdown import markdown_to_html5, markdown_to_html5_deep
from feedmark.utils import quote


def rewrite_reference_links(refdex, reference_links):
    new_reference_links = []
    seen_names = set()
    for (name, url) in reference_links:
        if name in seen_names:
            continue
        seen_names.add(name)
        if name in refdex:
            entry = refdex[name]
            if 'filename' in entry and 'anchor' in entry:
                filename = quote(entry['filename'].encode('utf-8'))
                anchor = quote(entry['anchor'].encode('utf-8'))
                url = u'{}#{}'.format(filename, anchor)
            elif 'filenames' in entry and 'anchor' in entry:
                # pick the last one, for compatibility with single-refdex style
                filename = quote(entry['filenames'][-1].encode('utf-8'))
                anchor = quote(entry['anchor'].encode('utf-8'))
                url = u'{}#{}'.format(filename, anchor)
            elif 'url' in entry:
                url = entry['url']
            else:
                raise ValueError("Badly formed refdex entry: {}".format(entry))
        new_reference_links.append((name, url))
    return new_reference_links


class Document(object):
    def __init__(self, title):
        self.title = title
        self.properties = OrderedDict()

        self.preamble = None
        self.sections = []

    def __str__(self):
        return "document '{}'".format(self.title.encode('utf-8'))

    def rewrite_reference_links(self, refdex):
        self.reference_links = rewrite_reference_links(refdex, self.reference_links)
        for section in self.sections:
            section.reference_links = rewrite_reference_links(refdex, section.reference_links)

    def global_reference_links(self):
        reference_links = []
        reference_links.extend(self.reference_links)
        for section in self.sections:
            reference_links.extend(section.reference_links)
        return reference_links

    def to_json_data(self, **kwargs):

        if kwargs.get('htmlize', False):
            if 'reference_links' not in kwargs:
                kwargs['reference_links'] = self.global_reference_links()
            preamble = markdown_to_html5(self.preamble, reference_links=kwargs['reference_links'])
            properties = markdown_to_html5_deep(self.properties, reference_links=kwargs['reference_links'])
        else:
            preamble = self.preamble
            properties = self.properties

        if kwargs.get('ordered', False):
            properties_list = []
            for key, value in properties.items():
                properties_list.append([key, value])
            properties = properties_list
        else:
            properties = dict(properties)

        return {
            'filename': self.filename,
            'title': self.title,
            'properties': properties,
            'preamble': preamble,
            'sections': [s.to_json_data(**kwargs) for s in self.sections],
        }


class Section(object):
    def __init__(self, title):
        self.document = None
        self.title = title
        self.properties = OrderedDict()

        self.lines = []

    def __str__(self):
        s = "section '{}'".format(self.title.encode('utf-8'))
        if self.document:
            s += " of " + str(self.document)
        return s

    @property
    def body(self):
        return '\n'.join(self.lines)

    @property
    def publication_date(self):
        formats = (
            "%b %d %Y %H:%M:%S",
            "%a, %d %b %Y %H:%M:%S GMT",
        )
        for format in formats:
            try:
                return datetime.strptime(self.properties['date'], format)
            except KeyError:
                raise KeyError("could not find 'date' on {}".format(self))
            except ValueError:
                pass
        raise NotImplementedError

    @property
    def anchor(self):
        return re.sub(
            r'[-\s]',
            '-',
            re.sub(r'[^\w\s-]', '', self.title).strip().lower()
        )

    def to_json_data(self, **kwargs):

        if kwargs.get('htmlize', False):
            body = markdown_to_html5(self.body, reference_links=kwargs['reference_links'])
            properties = markdown_to_html5_deep(self.properties, reference_links=kwargs['reference_links'])
        else:
            body = self.body
            properties = self.properties

        if kwargs.get('ordered', False):
            properties_list = []
            for key, value in properties.items():
                properties_list.append([key, value])
            properties = properties_list
        else:
            properties = dict(properties)

        return {
            'title': self.title,
            'anchor': self.anchor,
            'images': self.images,
            'properties': properties,
            'body': body,
        }


class Parser(object):

    def __init__(self, doc):
        self.lines = doc.split('\n')
        self.index = 0
        self.line = self.lines[self.index]
        self.index += 1

    def scan(self):
        self.line = self.lines[self.index]
        self.index += 1

    def eof(self):
        return self.index > len(self.lines) - 1

    def is_blank_line(self):
        return re.match(r'^\s*$', self.line)

    def is_image_line(self):
        return (
            re.match(r'^\!\[.*?\]\(.*?\)\s*$', self.line) or
            re.match(r'^\[\!\[.*?\]\(.*?\)\]\(.*?\)\s*$', self.line)
        )

    def is_property_line(self):
        return re.match(r'^\*\s+(.*?)\s*(\:|\@)\s*(.*?)\s*$', self.line)

    def is_heading_line(self):
        return re.match(r'^\#\#\#\s+(.*?)\s*$', self.line)

    def is_reference_link_line(self):
        return re.match(r'^\[(.*?)\]\:\s*(.*?)\s*$', self.line)

    def parse_document(self):
        # Feed       ::= :Title Properties Body {Section}.
        # Section    ::= {:Blank} :Heading {Image} Properties Body.
        # Properties ::= {:Blank | :Property}.
        # Body       ::= {:NonHeadingLine}.

        title = self.parse_title()
        document = Document(title)
        document.properties = self.parse_properties()
        lines, reference_links = self.parse_body()
        document.preamble = u'\n'.join(lines)
        document.reference_links = reference_links
        while not self.eof():
            section = self.parse_section()
            section.document = document
            document.sections.append(section)
        return document

    def parse_title(self):
        match = re.match(r'^\#\s+(.*?)\s*$', self.line)
        if match:
            title = match.group(1)
            self.scan()
            return title
        match = re.match(r'^\s*(.*?)\s*$', self.line)
        if match:
            title = match.group(1)
            self.scan()
            match = re.match(r'^\s*(\=+)\s*$', self.line)
            if match:
                self.scan()
                return title
        raise ValueError('Expected title')

    def parse_property(self):
        match = re.match(r'^\*\s+(.*?)\s*\@\s*(.*?)\s*$', self.line)
        if match:
            (key, val) = (match.group(1), match.group(2))
            self.scan()
            return ('@', key, val)
        match = re.match(r'^\*\s+(.*?)\s*\:\s*(.*?)\s*$', self.line)
        if match:
            (key, val) = (match.group(1), match.group(2))
            self.scan()
            return (':', key, val)
        raise ValueError('Expected property')

    def parse_properties(self):
        properties = OrderedDict()
        while self.is_blank_line() or self.is_property_line():
            if self.is_property_line():
                kind, key, val = self.parse_property()
                if kind == ':':
                    if key in properties:
                        raise KeyError('{} already given'.format(key))
                    properties[key] = val
                elif kind == '@':
                    properties.setdefault(key, []).append(val)
                else:
                    raise NotImplementedError(kind)
            else:
                self.scan()
        return properties

    def parse_section(self):
        while self.is_blank_line():
            self.scan()

        match = re.match(r'^\#\#\#\s+(.*?)\s*(\#\#\#)?\s*$', self.line)
        if not match:
            raise ValueError('Expected section, found "{}"'.format(self.line))

        section = Section(match.group(1))
        self.scan()
        section.images = self.parse_images()
        section.properties = self.parse_properties()
        lines, reference_links = self.parse_body()
        section.lines = lines
        section.reference_links = reference_links
        return section

    def parse_images(self):
        images = []
        while self.is_blank_line() or self.is_image_line():
            if self.is_image_line():
                match = re.match(r'^\!\[(.*?)\]\((.*?)\)\s*$', self.line)
                if match:
                    images.append({
                        'description': match.group(1),
                        'source': match.group(2),
                    })
                else:
                    match = re.match(r'^\[\!\[(.*?)\]\((.*?)\)\]\((.*?)\)\s*$', self.line)
                    images.append({
                        'description': match.group(1),
                        'source': match.group(2),
                        'link': match.group(3),
                    })
            self.scan()
        return images

    def parse_body(self):
        lines = []
        reference_links = []
        while not self.eof() and not self.is_heading_line() and not self.is_reference_link_line():
            lines.append(self.line)
            self.scan()
        while not self.eof() and (self.is_reference_link_line() or self.is_blank_line()):
            if self.is_reference_link_line():
                match = re.match(r'^\[(.*?)\]\:\s*(.*?)\s*$', self.line)
                if match:
                    reference_links.append((match.group(1), match.group(2)))
            self.scan()
        return (lines, reference_links)


if __name__ == "__main__":
    import doctest
    doctest.testmod()



# Feed-related, but Atom-independent, functions.

from feedmark.utils import quote_plus


def construct_entry_url(section):
    # Currently supports links to anchors generated by Github's Markdown renderer.

    if 'link-target-url' not in section.document.properties:
        return None

    return '{}#{}'.format(section.document.properties['link-target-url'], quote_plus(section.anchor))


def extract_feed_properties(document):
    properties = {}
    properties['title'] = document.title
    properties['author'] = document.properties['author']
    properties['url'] = document.properties['url']
    properties['link-target-url'] = document.properties.get('link-target-url')
    return properties


def extract_sections(documents):
    sections = []
    for document in documents:
        for section in document.sections:
            sections.append(section)
    sections.sort(key=lambda section: section.publication_date, reverse=True)
    return sections



from __future__ import absolute_import

from feedmark.formats.markdown import markdown_to_html5
from feedmark.utils import items


class Schema(object):
    def __init__(self, document):
        self.document = document
        self.property_rules = {}
        self.property_priority_order = []
        for section in self.document.sections:
            self.property_rules[section.title] = section
            self.property_priority_order.append(section.title)

    def check(self, section):
        results = []
        for key, value in items(section.properties):
            if key not in self.property_rules:
                results.append(['extra', key])
        for key, value in items(self.property_rules):
            optional = value.properties.get('optional', 'false') == 'true'
            if optional:
                continue
            if key not in section.properties:
                results.append(['missing', key])
        return results

    def check_documents(self, documents):
        results = []
        for document in documents:
            for section in document.sections:
                result = self.check(section)
                if result:
                    results.append({
                        'section': section.title,
                        'document': document.title,
                        'result': result
                    })
        return results

    def get_property_priority_order(self):
        return self.property_priority_order


def extract_links(html_text):
    from bs4 import BeautifulSoup

    links = []
    soup = BeautifulSoup(html_text, 'html.parser')
    for link in soup.find_all('a'):
        url = link.get('href')
        links.append(url)

    return links


def extract_links_from_documents(documents):
    links = []

    def make_link(url, section=None, **kwargs):
        link = {
            'url': url,
        }
        if section:
            link.update({
                'section': section.title,
                'document': section.document.title,
            })
        link.update(kwargs)
        return link

    def extend_links(section, md):
        links.extend([make_link(url, section=section) for url in extract_links(markdown_to_html5(md))])

    for document in documents:
        for name, url in document.reference_links:
            links.append(make_link(url, name=name))
        for section in document.sections:
            for (name, url) in section.images:
                links.append(make_link(url, section=section, name=name))
            for key, value in items(section.properties):
                if isinstance(value, list):
                    for subitem in value:
                        extend_links(section, subitem)
                else:
                    extend_links(section, value)
            for name, url in section.reference_links:
                links.append(make_link(url, section=section, name=name))
            extend_links(section, section.body)
    return links



from __future__ import absolute_import

from collections import OrderedDict
import re

from feedmark.utils import items_in_priority_order, unicode


def remove_outer_p(html):
    match = re.match(r'^\s*\<\s*p\s*\>\s*(.*?)\s*\<\s*/\s*p\s*\>\s*$', html)
    if match:
        html = match.group(1)
    return html


def markdown_to_html5(text, reference_links=None):
    """Canonical function used within `feedmark` to convert Markdown text to a HTML5 snippet."""
    from markdown import markdown

    if reference_links:
        text += markdownize_reference_links(reference_links)

    return markdown(text, extensions=['markdown.extensions.toc'])


def markdown_to_html5_deep(obj, **kwargs):
    if obj is None:
        return None
    elif isinstance(obj, OrderedDict):
        return OrderedDict((k, markdown_to_html5_deep(v, **kwargs)) for k, v in obj.items())
    elif isinstance(obj, dict):
        return dict((k, markdown_to_html5_deep(v, **kwargs)) for k, v in obj.items())
    elif isinstance(obj, list):
        return [markdown_to_html5_deep(subobj, **kwargs) for subobj in obj]
    else:
        return remove_outer_p(markdown_to_html5(unicode(obj), **kwargs))


def markdownize_properties(properties, property_priority_order):
    if not properties:
        return ''
    md = ''
    for key, value in items_in_priority_order(properties, property_priority_order):
        if isinstance(value, list):
            for subitem in value:
                md += u'*   {} @ {}\n'.format(key, subitem)
        else:
            md += u'*   {}: {}\n'.format(key, value)
    md += '\n'
    return md


def markdownize_reference_links(reference_links):
    if not reference_links:
        return u''
    md = u''
    md += u'\n'
    for name, url in reference_links:
        md += u'[{}]: {}\n'.format(name, url)
    return md


def feedmark_markdownize(document, schema=None):
    property_priority_order = []
    if schema is not None:
        property_priority_order = schema.get_property_priority_order()

    md = u'{}\n{}\n\n'.format(document.title, '=' * len(document.title))
    md += markdownize_properties(document.properties, property_priority_order)
    md += document.preamble
    md += markdownize_reference_links(document.reference_links)
    for section in document.sections:
        md += u'\n'
        md += u'### {}\n\n'.format(section.title)
        if section.images:
            for entry in section.images:
                if 'link' in entry:
                    md += u'[![{}]({})]({})\n'.format(
                        entry['description'], entry['source'], entry['link'],
                    )
                else:
                    md += u'![{}]({})\n'.format(
                        entry['description'], entry['source'],
                    )
            md += u'\n'
        md += markdownize_properties(section.properties, property_priority_order)
        md += section.body
        md += markdownize_reference_links(section.reference_links)
    md += '\n'
    return md


def feedmark_htmlize(document, *args, **kwargs):
    return markdown_to_html5(feedmark_markdownize(document, *args, **kwargs))
